package com.cirrial.finalbite {

import com.cirrial.finalbite.boss.Boss;
import com.cirrial.finalbite.effects.GiantExplodeEmitter;
import com.cirrial.finalbite.effects.LargeExplodeEmitter;
import com.cirrial.finalbite.effects.SmallExplodeEmitter;
import com.cirrial.finalbite.enemy.Enemy;
import com.cirrial.finalbite.enemy.ImpulseEnemy;
import com.cirrial.finalbite.enemy.ShooterEnemy;
import com.cirrial.finalbite.enemy.StandardEnemy;
import com.cirrial.finalbite.projectile.Projectile;
import com.cirrial.finalbite.widgets.FishPowerup;
import com.cirrial.finalbite.widgets.MantaWidget;
import com.cirrial.finalbite.widgets.RemoraPowerup;
import com.cirrial.util.G;

import flash.geom.Rectangle;

import org.flixel.*;
import org.flixel.plugin.photonstorm.FlxBar;

public class PlayState extends FlxState {
    [Embed(source='../../../../data/graphics/background.png')]
    protected var ImgBackdrop:Class;

    public static var player:Player;
    public static var boss:Boss;
    public static var manta:MantaWidget;
    public static var remora:Remora;
    public static var levelMap:FlxTilemap;
    public static var enemyGroup:FlxGroup;
    public static var powerupGroup:FlxGroup;
    public static var projectileGroup:FlxGroup;
    public static var effectGroup:FlxGroup;
    public static var hudGroup:FlxGroup;
    public static var limbGroup:FlxGroup;

    public static var healthBar:FlxBar;
    public static var enemyBar:FlxBar;
    public static var remoraBar:FlxBar;
    public static var playerNameText:FlxText;
    public static var enemyNameText:FlxText;
    public static var remoraNameText:FlxText;
    public static var escapeText:FlxText;

    public static var cameraFocus:FlxPoint;
    public static var cameraFocusDest:FlxPoint;

    public static var level:BaseLevel;

    public static var gameWon:Boolean = false;
    public static var advancementDisabled:Boolean = false;

    public static var lastEnemyStruck:Enemy;

    override public function create():void {
        lastEnemyStruck = null;
        limbGroup = new FlxGroup();
        enemyGroup = new FlxGroup();
        powerupGroup = new FlxGroup();
        projectileGroup = new FlxGroup();
        effectGroup = new FlxGroup();
        hudGroup = new FlxGroup();

        var backdrop:FlxSprite = new FlxSprite(0,0,ImgBackdrop);
        backdrop.scrollFactor.x = 0;
        backdrop.scrollFactor.y = 0;
        add(backdrop);

        level = new Level_Level1(true, onObjectAddedCallback);
        postLevelInit();

        initHud();

        add(limbGroup);
        add(projectileGroup);
        add(effectGroup);
        add(hudGroup);

        Main.framesTaken = 0;

        FlxG.flash(0xFF000000, 0.3);

        FlxG.playMusic(SoundLibrary.MusSpanishark, 0.8);

    }

    protected static function initHud():void {
        playerNameText = new FlxText(10, 10, G.SCREEN_WIDTH-20, Main.importCharName1);
        playerNameText.setFormat("gameFont", 24, 0xFFFFFFFF, "left", 0xFF204020);
        playerNameText.scrollFactor.x = 0.0;
        playerNameText.scrollFactor.y = 0.0;

        healthBar = new FlxBar(10, 50, FlxBar.FILL_LEFT_TO_RIGHT, 200, 20, player, "health", 0, player.health);
        healthBar.createFilledBar(0xFF208000, 0xFF00FF00, true, 0xFF000000);
        healthBar.killOnEmpty = true;
        healthBar.scrollFactor.x = 0.0;
        healthBar.scrollFactor.y = 0.0;

        remoraNameText = new FlxText(10, 90, G.SCREEN_WIDTH-20, Main.importCharName2);
        remoraNameText.setFormat("gameFont", 12, 0xFFFFFFFF, "left", 0xFF204020);
        remoraNameText.text = Main.importCharName2 + " is helping out!";
        remoraNameText.scrollFactor.x = 0.0;
        remoraNameText.scrollFactor.y = 0.0;

        remoraBar = new FlxBar(10, 120, FlxBar.FILL_LEFT_TO_RIGHT, 200, 20, remora, "helpTimer", 0, remora.helpDuration);
        remoraBar.createFilledBar(0xFF008040, 0xFF00FF80, true, 0xFF000000);
        remoraBar.scrollFactor.x = 0.0;
        remoraBar.scrollFactor.y = 0.0;

        enemyNameText = new FlxText(10, 10, G.SCREEN_WIDTH-20, Main.importCharName2);
        enemyNameText.setFormat("gameFont", 24, 0xFFFFFFFF, "right", 0xFF204020);
        enemyNameText.text = Main.importCharName2 + " is helping out!";
        enemyNameText.scrollFactor.x = 0.0;
        enemyNameText.scrollFactor.y = 0.0;

        enemyBar = new FlxBar(G.SCREEN_WIDTH - 210, 50, FlxBar.FILL_RIGHT_TO_LEFT, 200, 20, null, "", 0, 100);
        enemyBar.createFilledBar(0xFF800000, 0xFFFF2000, true, 0xFF000000);
        enemyBar.scrollFactor.x = 0.0;
        enemyBar.scrollFactor.y = 0.0;

        escapeText = new FlxText(10, 10, G.SCREEN_WIDTH-20, "!! Enemies Not Defeated !!");
        escapeText.setFormat("gameFont", 24, 0xFFFF3030, "center", 0xFF204020);
        escapeText.alpha = 0.0;
        escapeText.scrollFactor.x = 0.0;
        escapeText.scrollFactor.y = 0.0;

        hudGroup.add(healthBar);
        hudGroup.add(remoraBar);
        hudGroup.add(enemyBar);
        hudGroup.add(playerNameText);
        hudGroup.add(remoraNameText);
        hudGroup.add(enemyNameText);
        hudGroup.add(escapeText);
    }

    override public function update():void {
        super.update();

        var e:Enemy;

        // CAMERA CONTROL / WAVE ADVANCEMENT
        //////////////////////////////////////////////////////////
        // if player crosses boundaries, adjust camera
        var screenChanged:Boolean = false;
        if((player.x + player.width/2) >= cameraFocusDest.x + G.SCREEN_WIDTH/2) {
            // scroll right
            if(!advancementDisabled) {
                cameraFocusDest.x += G.SCREEN_WIDTH;
                screenChanged = true;
            } else {
                player.x = cameraFocusDest.x + G.SCREEN_WIDTH/2 - player.width/2;
                player.velocity.x = 0;
                escapeText.alpha = 1.0;
            }
        } else if((player.x + player.width/2) <= cameraFocusDest.x - G.SCREEN_WIDTH/2) {
            // scroll left
            if(!advancementDisabled) {
                cameraFocusDest.x -= G.SCREEN_WIDTH;
                screenChanged = true;
            } else {
                player.x = cameraFocusDest.x - G.SCREEN_WIDTH/2 - player.width/2;
                player.velocity.x = 0;
                escapeText.alpha = 1.0;
            }
        }

        // sleep/wake all enemies
        if(screenChanged) {
            for each(e in enemyGroup.members) {
                e.offScreenDisable();
                if((e.x < cameraFocusDest.x + G.SCREEN_WIDTH/2) && (e.x + e.width > cameraFocusDest.x - G.SCREEN_WIDTH/2)) {
                    advancementDisabled = true;
                    e.offScreenEnable();
                    if(e == boss) {
                        // play boss music
                        FlxG.playMusic(SoundLibrary.MusTowerOfPeeps);
                    }
                }
            }
        }

        cameraFocus.x += (cameraFocusDest.x - cameraFocus.x) * 0.1;
        FlxG.camera.focusOn(cameraFocus);
        //////////////////////////////////////////////////////////


        // COLLISION LOGIC
        //////////////////////////////////////////////////////////
        FlxG.collide(player, levelMap);
        FlxG.collide(enemyGroup, levelMap);
        FlxG.collide(projectileGroup, levelMap, killBullet);

        FlxG.overlap(player, enemyGroup, onPlayerEnemyOverlap);
        FlxG.overlap(player, projectileGroup, onPlayerProjectileOverlap);
        if(remora != null && remora.alive) FlxG.overlap(remora, enemyGroup, onAllyEnemyOverlap);
        FlxG.overlap(player, powerupGroup, onPlayerPowerupOverlap);
        //////////////////////////////////////////////////////////

        // GAME LOGIC
        //////////////////////////////////////////////////////////
        if(!player.alive) {
            FlxG.fade(0xFF000000, 0.3, onFade);
        }

        if(!boss.alive) {
            manta.play("safe");
        }

        // allow progression if all enemies defeated
        if(advancementDisabled) {
            var enemyOnScreen:Boolean = false;
            for each(e in enemyGroup.members) {
                if(e.isOffScreenEnabled() && e.alive) {
                    enemyOnScreen = true;
                }
            }
            if(!enemyOnScreen) advancementDisabled = false;
        }
        //////////////////////////////////////////////////////////

        // HUD AND MISC LOGIC
        //////////////////////////////////////////////////////////

        if(remora.alive) {
            remoraBar.visible = true;
            remoraNameText.visible = true;
        } else {
            remoraBar.visible = false;
            remoraNameText.visible = false;
        }

        if(lastEnemyStruck != null && lastEnemyStruck.alive) {
            enemyNameText.text = lastEnemyStruck.name;
            enemyBar.setParent(lastEnemyStruck, "health");
            enemyNameText.visible = true;
            enemyBar.visible = true;
        } else {
            enemyNameText.visible = false;
            enemyBar.visible = false;
        }

        // do not allow enemies to leave the screen
        for each(var enemy:Enemy in enemyGroup.members) {
            if(enemy.isOffScreenEnabled()) {
                if((enemy.x + enemy.width/2) >= cameraFocusDest.x + G.SCREEN_WIDTH/2) {
                    enemy.x = cameraFocusDest.x + G.SCREEN_WIDTH/2 - enemy.width/2;
                    enemy.velocity.x = 0;
                } else if((enemy.x + enemy.width/2) <= cameraFocusDest.x - G.SCREEN_WIDTH/2) {
                    enemy.x = cameraFocusDest.x - G.SCREEN_WIDTH/2 - enemy.width/2;
                    enemy.velocity.x = 0;
                }

                if(enemy.y > G.SCREEN_HEIGHT) enemy.y = G.SCREEN_HEIGHT - 100;
                if(enemy.x > level.boundsMax.x) enemy.x = level.boundsMax.x - 100;
            }
        }


        // TERRIBLE HACKY FIX TO REPOSITION PLAYER IF THEY FALL OUT OF BOUNDS

        if(player.y > G.SCREEN_HEIGHT) player.y = G.SCREEN_HEIGHT - 100;
        if(player.x > level.boundsMax.x) player.x = level.boundsMax.x - 100;

        Main.framesTaken++;
        if(escapeText.alpha >= 0) escapeText.alpha -= 0.02;
        //////////////////////////////////////////////////////////
    }

    protected static function killBullet(p:Projectile, m:FlxTilemap):void {
        p.kill();
    }

    protected static function onPlayerProjectileOverlap(sprite1:FlxSprite, sprite2:FlxSprite):void {
        if(sprite1 == player) {
            var bullet:Projectile = sprite2 as Projectile;
            var dir:FlxPoint = G.getDirToObject(bullet, player);
            bullet.kill();
            if(player.isVulnerable()) {
                player.hurt(bullet.damage);
                player.velocity.x = dir.x * bullet.force;
                player.velocity.y = dir.y * bullet.force;
            }
        }
    }

    protected static function onAllyEnemyOverlap(sprite1:FlxSprite, sprite2:FlxSprite):void {
        if(sprite1 is Remora) {
            var remora:Remora = sprite1 as Remora;
            var enemy:Enemy = sprite2 as Enemy;
            var dir:FlxPoint = G.getDirToObject(enemy, player);
            if(remora.isAttacking() && enemy.isVulnerable()) {
                enemy.hurt(remora.attackDamage);
                enemy.velocity.x = dir.x * remora.attackForce;
                enemy.velocity.y = dir.y * remora.attackForce;
            }
        }
    }

    protected function onPlayerPowerupOverlap(sprite1:FlxSprite, sprite2:FlxSprite):void {
        // safely assume that player is the thing being tested
        if(sprite1 == player) {
            if(sprite2 is FishPowerup && player.health < 100) {
                var fish:FishPowerup = sprite2 as FishPowerup;
                fish.kill();
                player.health += fish.healthBoost;
                FlxG.flash(0xFF20FFA0, 0.1);
                G.playSound(SoundLibrary.SndGetFish, fish);
            } else if(sprite2 is RemoraPowerup) {
                if(!remora.alive) {
                    sprite2.kill();
                    remora.reset(sprite2.x, sprite2.y);
                    G.playSound(SoundLibrary.SndGetRemora, remora);
                }
            } else if (sprite2 is MantaWidget) {
                if(!boss.alive && !gameWon) {
                    gameWon = true;
                    FlxG.fade(0xFFFFFFFF, 1.0, onFade);
                }
            }
        }
    }

    private static function onFade():void {
        // MOVE TO THE ENDING OF THE GAME!!
        if(gameWon) {
            FlxG.music.stop();
            FlxG.switchState(new EndState());
        } else {
            // Restart the game!
            FlxG.switchState(new PlayState());
        }
    }

    protected static function onPlayerEnemyOverlap(sprite1:FlxSprite, sprite2:FlxSprite):void {
        // safely assume that player is the thing being tested
        if(sprite1 == player) {
            if(sprite2 is Enemy) {
                var enemy:Enemy = sprite2 as Enemy;
                var dir:FlxPoint = G.getDirToObject(enemy, player);
                if(player.isAttacking() && enemy.isVulnerable()) {
                    if(player.isBiting() && player.hasNoBiteTarget()) {
                        player.bite(enemy);
                    } else {
                        enemy.hurt(player.attackDamage);
                        enemy.velocity.x = dir.x * player.attackForce;
                        enemy.velocity.y = dir.y * player.attackForce;
                        lastEnemyStruck = enemy;
                    }
                } else if(enemy.isAttacking() && player.isVulnerable()) {
                    player.hurt(enemy.attackDamage);
                    player.velocity.x = -dir.x * enemy.attackForce;
                    player.velocity.y = -dir.y * enemy.attackForce;
                }
            }
        }
    }

    public function onObjectAddedCallback(obj:Object, group:FlxGroup, lvl:BaseLevel, scrollX:Number, scrollY:Number, properties:Array):Object {
        if(obj is Player) {
            player = obj as Player;
        } else if(obj is Boss) {
            boss = obj as Boss;
            boss.createLimbs(enemyGroup, limbGroup);
            boss.offScreenDisable();
            enemyGroup.add(boss);
        } else if (obj is Enemy) {
            var enemy:Enemy = obj as Enemy;
            enemy.offScreenDisable();
            enemyGroup.add(obj as Enemy);
        }else if (obj is FishPowerup) {
            powerupGroup.add(obj as FishPowerup);
        } else if (obj is RemoraPowerup) {
            powerupGroup.add(obj as RemoraPowerup);
        } else if (obj is MantaWidget) {
            manta = obj as MantaWidget;
            powerupGroup.add(obj as MantaWidget);
        } else if (obj is FlxTilemap) {
            levelMap = obj as FlxTilemap;
        }
        return obj;
    }

    public function postLevelInit():void {
        FlxG.worldBounds.make(level.boundsMin.x, level.boundsMin.y, level.boundsMax.x, level.boundsMax.y);
        boss.addLimbEndsToState(this);

        var initCameraX:Number = G.SCREEN_WIDTH * (Math.floor(player.x/G.SCREEN_WIDTH) + 0.5);
        var initCameraY:Number = G.SCREEN_HEIGHT/2;
        cameraFocus = new FlxPoint(initCameraX, initCameraY);
        cameraFocusDest = new FlxPoint(initCameraX, initCameraY);

        for each(var e:Enemy in enemyGroup.members) {
            e.setTarget(player);
            if(FlxU.getDistance(cameraFocusDest, new FlxPoint(e.x+ e.width/2, e.y+ e.height/2)) <= G.SCREEN_WIDTH/2) {
                e.offScreenEnable();
                advancementDisabled = true;
            }
        }

        remora = new Remora(0,0);
        remora.kill();
        add(remora);
    }

    public static function smallExplode(source:FlxSprite):void {
        var explodeEmit:SmallExplodeEmitter = effectGroup.recycle(SmallExplodeEmitter) as SmallExplodeEmitter;
        //G.playSound(SoundLibrary.SndEnemyDie, source);
        explodeEmit.boom(source.x + source.width/2, source.y + source.height/2);
    }

    public static function largeExplode(source:FlxSprite):void {
        var explodeEmit:LargeExplodeEmitter = effectGroup.recycle(LargeExplodeEmitter) as LargeExplodeEmitter;
        //G.playSound(SoundLibrary.SndEnemyDie, source);
        explodeEmit.boom(source.x + source.width/2, source.y + source.height/2);
    }

    public static function giantExplode(source:FlxSprite):void {
        var explodeEmit:GiantExplodeEmitter = effectGroup.recycle(GiantExplodeEmitter) as GiantExplodeEmitter;
        //G.playSound(SoundLibrary.SndEnemyDie, source);
        explodeEmit.boom(source.x + source.width/2, source.y + source.height/2);
    }
}
}